import { Aragami, ClassType } from 'aragami';
import { ConsoleLogger, Logger } from '@nestjs/common';
import * as os from 'os';
import { InjectAragami } from '../index';

export interface QueueRunnerOptions {
  maxConcurrency?: number;
  key?: string;
  logTask?: boolean;
}

export class _QueueRunner<T> {
  logger = new Logger(this._queueClass.name);
  _maxConcurrency = this._queueRunnerOptions.maxConcurrency ?? os.cpus().length;

  constructor(
    public _queueClass: ClassType<T>,
    public _queueRunnerOptions: QueueRunnerOptions,
  ) {}

  @InjectAragami()
  aragami: Aragami;

  async onApplicationBootstrap() {
    if (this._queueRunnerOptions.logTask) {
      this.logger.log(`Starting ${this._maxConcurrency} workers`);
    }
    for (let i = 0; i < this._maxConcurrency; i++) {
      if (this._queueRunnerOptions.logTask) {
        this.logger.log(`Starting worker ${i}`);
      }
      this._mainLoop(i).then();
    }
  }

  _quit = false;

  async onApplicationShutdown() {
    this._quit = true;
  }

  async runTask(task: T, i: number) {
    return;
  }

  async _mainLoop(i: number) {
    const logger = new ConsoleLogger(`${this._queueClass.name}-${i}`);
    while (!this._quit) {
      logger.log('Looping');
      try {
        const task: T = await this.aragami.queueGatherBlocking(
          this._queueClass,
          this._queueRunnerOptions.key || 'default',
        );
        if (!task) continue;
        if (this._queueRunnerOptions.logTask) {
          logger.log(`Got task: ${JSON.stringify(task)}`);
        }
        await this.runTask(task, i);
      } catch (e) {
        logger.error(`Loop failed: ${e}`);
      }
    }
  }
}

export function QueueRunner<T>(
  queueClass: ClassType<T>,
  options?: Partial<QueueRunnerOptions>,
) {
  const cl = class extends _QueueRunner<T> {
    constructor(innerOptions?: Partial<QueueRunnerOptions>) {
      super(queueClass, {
        ...(innerOptions || {}),
        ...(options || {}),
      });
    }
  };
  Object.defineProperty(cl, 'name', { value: queueClass.name + 'Runner' });
  return cl;
}
